import {NS} from "@ns"
import { parse } from "./lib/flag";
import {range} from "./util/range";

class Account{
    value: number;
    timer?: NodeJS.Timer;
    constructor(value: number){
        this.value = value;
        this.timer = undefined;
    }
    startIncrement(ns:NS, rate: number): void{
        this.timer = setInterval(()=>{
            this.value += ns.hacknet.getNodeStats().production * rate;
        },1000);
    }
    stopIncrement(): void{
        clearInterval(this.timer)
    }
}

export function getNodeProduction(level: number, ram: number, cores: number, mult: number, bitMult:number): number{
    const moneyGainPerLevel = 1.5;
    const levelMult = level * moneyGainPerLevel;
    const ramMult = Math.pow(1.035, ram - 1);
    const coreMult = ((cores + 5) / 6);
    return levelMult * ramMult * coreMult * mult * bitMult;
}

interface HacknetNodeComponent<T> {
    level: T;
    core: T;
    ram: T;
}

type ComponentKind = "level" | "core" | "ram";

export interface CEInfo {
    upgradeCost : HacknetNodeComponent<number>;
    production : HacknetNodeComponent<number>;
    costEffective : HacknetNodeComponent<number>;
}

function getCostEffective(ns:NS, nodeIndex: number, n:number): CEInfo{
    const coreUpgradeCost = ns.hacknet.getCoreUpgradeCost(nodeIndex,n);
    const ramUpgradeCost = ns.hacknet.getRamUpgradeCost(nodeIndex,n);
    const levelUpgradeCost = ns.hacknet.getLevelUpgradeCost(nodeIndex,n);
    const mult = ns.getHacknetMultipliers().production;
    const bitMult = 1//ns.getBitNodeMultipliers().HacknetNodeMoney;
    const stat = ns.hacknet.getNodeStats(nodeIndex);
    
    const coreProduction = getNodeProduction(stat.level, stat.ram, stat.cores + n, mult, bitMult) - stat.production;
    const levelProduction = getNodeProduction(stat.level + n, stat.ram, stat.cores, mult, bitMult) - stat.production;
    const ramProduction = getNodeProduction(stat.level, stat.ram * Math.pow(2,n), stat.cores, mult, bitMult) - stat.production;

    return {
        upgradeCost:{
            core: coreUpgradeCost,
            ram: ramUpgradeCost,
            level: levelUpgradeCost,
        },
        production:{
            core: coreProduction,
            ram: ramProduction,
            level: levelProduction,
        },
        costEffective:{
            core: coreProduction/coreUpgradeCost,
            ram: ramProduction/ramUpgradeCost,
            level: levelProduction/levelUpgradeCost
        }
    }
}

function createUpgradeAction(ns:NS, index: number, n: number, kind:ComponentKind):(account: Account)=>void{
    return (account: Account)=>{
        switch (kind) {
            case "core":
                account.value -= ns.hacknet.getCoreUpgradeCost(index,n);
                ns.hacknet.upgradeCore(index,n)
                break;
            case "level":
                account.value -= ns.hacknet.getLevelUpgradeCost(index,n);
                ns.hacknet.upgradeLevel(index,n)
                break;
            case "ram":
                account.value -= ns.hacknet.getRamUpgradeCost(index,n);
                ns.hacknet.upgradeRam(index,n)
                break;
        }
    }
}

function getMinCENode(ns:NS, index: number, n: number):[number,ComponentKind]{
    const {costEffective: ce} = getCostEffective(ns,index,n);
    if(ce.core < ce.level){
        if(ce.core < ce.ram){
            return [ce.core, "core"];
        }
        else{
            return [ce.ram, "ram"];
        }
    }
    else {
        if(ce.level < ce.ram){
            return [ce.level,"level"];
        }
        else {
            return [ce.ram,"ram"];
        }
    }
}

function* getIndexOfHackNode(ns:NS): Generator<number>{
    for (const it of range(ns.hacknet.numNodes())) {
        yield it;   
    }
}

function getMinCEIndex(ns:NS, n: number):[number,number,ComponentKind]{
    let optimalC = Infinity;
    let kind = "";
    let minIndex = 0;
    for(const it of getIndexOfHackNode(ns)){
        const [c, k] = getMinCENode(ns,it,n);
        if(c < optimalC){
            optimalC = c;
            kind = k;
            minIndex = it;
        }
    }
    return [minIndex,optimalC,kind];
}

function getBuyNodeCE(ns:NS):{cost:number;production:number;costEffective: number}{
    const mult = ns.getHacknetMultipliers().production;
    const bitMult = 1//ns.getBitNodeMultipliers().HacknetNodeMoney;
    const cost = ns.hacknet.maxNumNodes() > ns.hacknet.numNodes() ? ns.hacknet.getPurchaseNodeCost() : Infinity;
    const production = getNodeProduction(1,1,1,mult,bitMult);
    return {
        cost,
        production,
        costEffective: cost/production,
    };
}

function cycle(ns:NS, account:Account): void{
    const chunkN = 1;

    const [upgradeNodeIndex,upgradeNodeCE,kind] = getMinCEIndex(ns,chunkN);
    const {costEffective: buyNew} = getBuyNodeCE(ns);
    if(upgradeNodeCE < buyNew){
        const action = createUpgradeAction(ns,upgradeNodeIndex,chunkN,kind);
        action(account);
    }
    else {
        ns.hacknet.purchaseNode()
    }
}

// eslint-disable-next-line require-await
export async function main(ns:NS):Promise<void>{
    ns.disableLog("ALL")
    const flag = parse(ns.args);
    const iter = parseInt(flag.i ?? flag.iter ?? "10");
    const rate = parseFloat(flag.r ?? flag.rate ?? "0.5");
    const budget = parseInt(flag.b ?? flag.budget ?? "0");

    const account = new Account(budget);
    account.startIncrement(ns,rate);
    ns.tail();
    for (let i = 0; i < iter; i++) {
        ns.clearLog()
        ns.print(`budget ${account.value}`);
        cycle(ns,account);
        await ns.sleep(1000);
    }
}